# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

import matplotlib.pyplot as plt
import numpy as np

from bezier import _intersection_helpers

import runtime_utils


CONFIG = runtime_utils.Config()
# Always gives us the unit square.
UNIT_SQUARE = np.array([
    [0.0, 0.0],
    [1.0, 1.0],
])


def make_plot(segment):
    figure = plt.figure()
    ax = figure.gca()

    line, = ax.plot([0, 1], [0, 1], alpha=0.0)
    ax.fill_between([0, 1], [0, 0], [1, 1],
                    alpha=0.5, color=line.get_color())
    line, = ax.plot(segment[:, 0], segment[:, 1])
    ax.plot(segment[0, 0], segment[0, 1], marker='o',
            linestyle='None', color=line.get_color())
    left_, bottom_ = np.min(segment, axis=0)
    right_, top_ = np.max(segment, axis=0)
    ax.fill_between([left_, right_], [bottom_, bottom_], [top_, top_],
                    alpha=0.5, color=line.get_color())
    ax.axis('scaled')
    ax.set_xlim(-1.125, 2.125)
    ax.set_ylim(-1.125, 2.125)

    if CONFIG.save_plot:
        CONFIG.save_fig()
    else:
        plt.title(CONFIG.current_test)
        plt.show()

    plt.close(figure)


def run_it(segment, expected=None):
    if expected is None:
        expected = _intersection_helpers.BoxIntersectionType.intersection

    result = _intersection_helpers.bbox_line_intersect(
        UNIT_SQUARE, segment[[0], :], segment[[1], :])
    assert result is expected

    if not CONFIG.running:
        return

    make_plot(segment)


def test_outside():
    segment_bottom_left = np.array([
        [0.25, -0.75],
        [-0.75, 0.25],
    ])
    segment_bottom_right = np.array([
        [0.75, -0.75],
        [1.75, 0.25],
    ])
    segment_top_right = np.array([
        [0.75, 1.75],
        [1.75, 0.75],
    ])
    segment_top_left = np.array([
        [-0.75, 0.75],
        [0.25, 1.75],
    ])

    segments = (
        segment_bottom_left,
        segment_bottom_right,
        segment_top_right,
        segment_top_left,
    )
    expected = _intersection_helpers.BoxIntersectionType.disjoint
    for segment in segments:
        run_it(segment, expected)


def test_start_in_box():
    segments = (
        np.array([
            [0.5, 0.0],
            [0.5, -0.5],
        ]),
        np.array([
            [0.5, 0.0],
            [0.5, 0.5],
        ]),
        np.array([
            [1.0, 0.5],
            [1.5, 0.5],
        ]),
        np.array([
            [1.0, 0.5],
            [0.5, 0.5],
        ]),
        np.array([
            [0.5, 1.0],
            [0.5, 1.5],
        ]),
        np.array([
            [0.5, 1.0],
            [0.5, 0.5],
        ]),
        np.array([
            [0.0, 0.5],
            [-0.5, 0.5],
        ]),
        np.array([
            [0.0, 0.5],
            [0.5, 0.5],
        ]),
        np.array([
            [0.5, 0.5],
            [1.5, 0.5],
        ]),
        np.array([
            [0.5, 0.5],
            [0.5, -0.5],
        ]),
        np.array([
            [0.5, 0.5],
            [-0.5, 1.5],
        ]),
    )

    for segment in segments:
        run_it(segment)


def test_end_in_box():
    segments = (
        np.array([
            [0.5, -1.0],
            [0.5, 0.0],
        ]),
        np.array([
            [0.25, -1.0],
            [0.5, 0.0],
        ]),
        np.array([
            [1.5, 0.5],
            [1.0, 0.5],
        ]),
        np.array([
            [1.0, -0.5],
            [1.0, 0.5],
        ]),
        np.array([
            [0.5, 1.5],
            [0.5, 1.0],
        ]),
        np.array([
            [0.5, -0.5],
            [0.5, 1.0],
        ]),
        np.array([
            [-0.5, 0.5],
            [0.0, 0.5],
        ]),
        np.array([
            [0.5, 1.5],
            [0.5, 0.5],
        ]),
        np.array([
            [-0.5, 0.5],
            [0.5, 0.5],
        ]),
        np.array([
            [1.5, 1.5],
            [0.5, 0.5],
        ]),
    )

    for segment in segments:
        run_it(segment)


def test_goes_through_box():
    segments = (
        np.array([
            [0.5, -0.25],
            [1.25, 0.5],
        ]),
        np.array([
            [-0.25, 0.5],
            [0.5, -0.25],
        ]),
        np.array([
            [1.25, 0.5],
            [0.5, 1.25],
        ]),
        np.array([
            [0.5, 1.25],
            [-0.25, 0.5],
        ]),
        np.array([
            [-0.5, 0.5],
            [1.5, 0.5],
        ]),
        np.array([
            [-0.25, 0.0],
            [1.25, 0.0],
        ]),
        np.array([
            [1.0, -0.25],
            [1.0, 1.25],
        ]),
        np.array([
            [1.25, 1.0],
            [-0.25, 1.0],
        ]),
        np.array([
            [0.0, 1.25],
            [0.0, -0.25],
        ]),
    )

    for segment in segments:
        run_it(segment)


if __name__ == '__main__':
    CONFIG.run(globals())



# See the License for the specific language governing permissions and
# limitations under the License.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import numpy as np

from tests import utils


class TestCurvedPolygon(utils.NumPyTestCase):

    NODES0 = np.array([
        [0.0, 0.0],
        [0.5, -1.0],
        [1.0, 0.0],
    ])
    NODES1 = np.array([
        [1.0, 0.0],
        [0.5, 1.0],
        [0.0, 0.0],
    ])
    COLOR = (0.125, 0.125, 0.0)

    @staticmethod
    def _get_target_class():
        from bezier import curved_polygon

        return curved_polygon.CurvedPolygon

    def _make_one(self, *args, **kwargs):
        klass = self._get_target_class()
        return klass(*args, **kwargs)

    def _make_default(self):
        import bezier

        edge0 = bezier.Curve(self.NODES0)
        edge1 = bezier.Curve(self.NODES1)
        return self._make_one(edge0, edge1)

    def test_constructor(self):
        import bezier

        edge0 = bezier.Curve(self.NODES0)
        edge1 = bezier.Curve(self.NODES1)
        curved_poly = self._make_one(edge0, edge1)
        self.assertEqual(curved_poly._edges, (edge0, edge1))
        self.assertEqual(curved_poly._num_sides, 2)

    def test__verify_too_few(self):
        with self.assertRaises(ValueError):
            self._make_one()
        with self.assertRaises(ValueError):
            self._make_one(None)

    def test__verify_bad_dimension(self):
        import bezier

        edge0 = bezier.Curve(np.array([
            [1.0, 1.0],
            [2.0, 2.0],
        ]))
        edge1 = bezier.Curve(self.NODES1)
        with self.assertRaises(ValueError):
            self._make_one(edge0, edge1)

    def test__verify_not_aligned(self):
        import bezier

        edge0 = bezier.Curve(np.array([[0.0], [0.0]]))
        edge1 = bezier.Curve(self.NODES1)
        with self.assertRaises(ValueError):
            self._make_one(edge0, edge1)

    def test_num_sides_property(self):
        curved_poly = self._make_default()
        self.assertIs(curved_poly.num_sides, 2)

    def test___repr__(self):
        curved_poly = self._make_default()
        self.assertEqual(repr(curved_poly),
                         '<CurvedPolygon (num_sides=2)>')

    def _check_plot_calls(self, ax):
        # Check the calls to ax.plot(). We can't assert_any_call()
        # since == breaks on NumPy arrays.
        self.assertEqual(ax.plot.call_count, 2)

        calls = ax.plot.mock_calls
        utils.check_plot_call(
            self, calls[0], self.NODES0[(0, 2), :], color=None)
        utils.check_plot_call(
            self, calls[1], self.NODES1[(0, 2), :], color=self.COLOR)

    def _make_axis(self):
        import matplotlib.lines

        ax = mock.Mock()

        line = matplotlib.lines.Line2D([], [], color=self.COLOR)
        ax.plot.return_value = (line,)

        return ax

    def _plot_helper(self, show=False, ax=None):
        if ax is None:
            orig_ax = False
            ax = self._make_axis()
        else:
            orig_ax = True

        curved_poly = self._make_default()
        plt = mock.Mock()

        figure = mock.Mock()
        plt.figure.return_value = figure
        figure.gca.return_value = ax

        with mock.patch('bezier.curved_polygon.plt', new=plt):
            kwargs = {}
            if show:
                kwargs['show'] = True
            if orig_ax:
                kwargs['ax'] = ax
            result = curved_poly.plot(2, **kwargs)

        self.assertIs(result, ax)

        # Check mocks.
        if orig_ax:
            plt.figure.assert_not_called()
            figure.gca.assert_not_called()
        else:
            plt.figure.assert_called_once_with()
            figure.gca.assert_called_once_with()

        ax.add_patch.assert_called_once()

        self._check_plot_calls(ax)

        if show:
            plt.show.assert_called_once_with()
        else:
            plt.show.assert_not_called()

    def test_plot(self):
        self._plot_helper()

    def test_plot_show(self):
        self._plot_helper(show=True)

    def test_plot_existing_axis(self):
        ax = self._make_axis()
        self._plot_helper(ax=ax)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

import matplotlib.pyplot as plt
import numpy as np
import pytest
import six

import bezier
from bezier import _intersection_helpers
from bezier import _plot_helpers

import runtime_utils


CONFIG = runtime_utils.Config()
# g1 = sympy.Matrix([[s, 2 * s * (1 - s)]])
CURVE1 = bezier.Curve(np.array([
    [0.0, 0.0],
    [0.5, 1.0],
    [1.0, 0.0],
]))
# g2 = sympy.Matrix([[(9 - 8 * s) / 8, (2 * s - 1)**2 / 2]])
CURVE2 = bezier.Curve(np.array([
    [1.125, 0.5],
    [0.625, -0.5],
    [0.125, 0.5],
]))
# g3 = 3 * g1
# g3 = sympy.Matrix([[3 * s, 6 * s * (1 - s)]])
CURVE3 = bezier.Curve(np.array([
    [0.0, 0.0],
    [1.5, 3.0],
    [3.0, 0.0],
]))
# g4 = sympy.Matrix([[
#     -3 * (4 * s**2 + s - 4) / 4,
#     (92 * s**2 - 77 * s + 24) / 16,
# ]])
CURVE4 = bezier.Curve(np.array([
    [3.0, 1.5],
    [2.625, -0.90625],
    [-0.75, 2.4375],
]))
# g5 = sympy.Matrix([[s, (8 * s**2 - 8 * s + 3) / 4]])
CURVE5 = bezier.Curve(np.array([
    [0.0, 0.75],
    [0.5, -0.25],
    [1.0, 0.75],
]))
# g6 = sympy.Matrix([[s, s**2 + (1 - s)**2]])
CURVE6 = bezier.Curve(np.array([
    [0.0, 1.0],
    [0.5, 0.0],
    [1.0, 1.0],
]))
# g7 = sympy.Matrix([[s, (4 * s**2 - 4 * s + 17) / 64]])
CURVE7 = bezier.Curve(np.array([
    [0.0, 0.265625],
    [0.5, 0.234375],
    [1.0, 0.265625],
]))
# g8 = sympy.Matrix([[8 * s, 3]]) / 8
CURVE8 = bezier.Curve(np.array([
    [0.0, 0.375],
    [1.0, 0.375],
]))
# g9 = sympy.Matrix([[2, 3 * s]]) / 4
CURVE9 = bezier.Curve(np.array([
    [0.5, 0.0],
    [0.5, 0.75],
]))
# g10 = 9 * g1
# g10 = sympy.Matrix([[9 * s, 18 * s * (1 - s)]])
CURVE10 = bezier.Curve(np.array([
    [0.0, 0.0],
    [4.5, 9.0],
    [9.0, 0.0],
]))
# g11 = sympy.Matrix([[6 * s, 8 * (1 - s)]])
CURVE11 = bezier.Curve(np.array([
    [0.0, 8.0],
    [6.0, 0.0],
]))
# NOTE: This curve has a self-crossing.
# g12 = sympy.Matrix([[
#     -3 * s * (3 * s - 2)**2 / 4,
#     -(27 * s**3 - 72 * s**2 + 48 * s - 16) / 8,
# ]])
CURVE12 = bezier.Curve(np.array([
    [0.0, 2.0],
    [-1.0, 0.0],
    [1.0, 1.0],
    [-0.75, 1.625],
]))
# g13 = sympy.Matrix([[s, 4 * s * (1 - s) * (7 * s**2 - 7 * s + 2)]])
CURVE13 = bezier.Curve(np.array([
    [0.0, 0.0],
    [0.25, 2.0],
    [0.5, -2.0],
    [0.75, 2.0],
    [1.0, 0.0],
]))
# g14 = sympy.Matrix([[3 * s / 4, 3 * s * (4 - 3 * s) / 8]])
CURVE14 = bezier.Curve(np.array([
    [0.0, 0.0],
    [0.375, 0.75],
    [0.75, 0.375],
]))
# g15 = sympy.Matrix([[(3 * s + 1) / 4, (9 * s**2 - 6 * s + 5) / 8]])
CURVE15 = bezier.Curve(np.array([
    [0.25, 0.625],
    [0.625, 0.25],
    [1.0, 1.0],
]))
# g16 = sympy.Matrix([[(3 * s + 1) / 4, 3 * (6 * s**2 - 4 * s + 3) / 16]])
CURVE16 = bezier.Curve(np.array([
    [0.25, 0.5625],
    [0.625, 0.1875],
    [1.0, 0.9375],
]))
# g17 = sympy.Matrix([[11 - 8 * s, -4 * (2 * s**2 - s - 2)]])
CURVE17 = bezier.Curve(np.array([
    [11.0, 8.0],
    [7.0, 10.0],
    [3.0, 4.0],
]))
# g18 = sympy.Matrix([[s + 1, -2 * s * (1 - s)]])
CURVE18 = bezier.Curve(np.array([
    [1.0, 0.0],
    [1.5, -1.0],
    [2.0, 0.0],
]))
# g19 = sympy.Matrix([[s + 1, 2 * s * (1 - s)]])
CURVE19 = bezier.Curve(np.array([
    [2.0, 0.0],
    [1.5, 1.0],
    [1.0, 0.0],
]))
# g20 = sympy.Matrix([[(2 * s - 1)**2, s/2]])
CURVE20 = bezier.Curve(np.array([
    [1.0, 0.0],
    [-1.0, 0.25],
    [1.0, 0.5],
]))
# g21 = sympy.Matrix([[
#     (10 * s - 1) / 8,
#     (9 - 10 * s) * (10 * s - 1) / 32,
# ]])
CURVE21 = bezier.Curve(np.array([
    [-0.125, -0.28125],
    [0.5, 1.28125],
    [1.125, -0.28125],
]))
# g22 = sympy.Matrix([[25 * (2 * s - 1)**2 / 16, (10 * s - 1)  / 16]])
CURVE22 = bezier.Curve(np.array([
    [1.5625, -0.0625],
    [-1.5625, 0.25],
    [1.5625, 0.5625],
]))
# g23 = 5 * g8 + sympy.Matrix([[24, 21]]) / 8
# g23 = sympy.Matrix([[10 * s + 6, 9]]) / 2
CURVE23 = bezier.Curve(np.array([
    [3.0, 4.5],
    [8.0, 4.5],
]))
# g24 = sympy.Matrix([[(4 * s + 1) / 4, (3 - 4 * s) * (4 * s + 1) / 8]])
CURVE24 = bezier.Curve(np.array([
    [0.25, 0.375],
    [0.75, 0.875],
    [1.25, -0.625],
]))
# g25 = sympy.Matrix([[
#     -s * (2 * s**2 - 3 * s - 3) / 4,
#     -(3 * s**3 - 3 * s - 1) / 2,
# ]])
CURVE25 = bezier.Curve(np.array([
    [0.0, 0.5],
    [0.25, 1.0],
    [0.75, 1.5],
    [1.0, 0.5],
]))
# g26 = sympy.Matrix([[
#     3 * (14 * s + 1) / 8,
#     18 * s**3 - 27 * s**2 + 3 * s + 7,
# ]])
CURVE26 = bezier.Curve(np.array([
    [0.375, 7.0],
    [2.125, 8.0],
    [3.875, 0.0],
    [5.625, 1.0],
]))
# g27 = sympy.Matrix([[
#     (6 * s + 1) / 8,
#     (35 * s**3 - 60 * s**2 + 24 * s + 4) / 16,
# ]])
CURVE27 = bezier.Curve(np.array([
    [0.125, 0.25],
    [0.375, 0.75],
    [0.625, 0.0],
    [0.875, 0.1875],
]))
# Rotate g1 by 45 degrees and scale by sqrt(2).
# g28 = sympy.Matrix([[s * (2 * s - 1), s * (3 - 2 * s)]])
CURVE28 = bezier.Curve(np.array([
    [0.0, 0.0],
    [-0.5, 1.5],
    [1.0, 1.0],
]))
# Rotate g6 by 45 degrees and scale by sqrt(2).
# g29 = sympy.Matrix([[(1 - 2 * s) * (1 - s), 2 * s**2 - s + 1]])
CURVE29 = bezier.Curve(np.array([
    [-1.0, 1.0],
    [0.5, 0.5],
    [0.0, 2.0],
]))
# Rotate g9 by 45 degrees and scale by sqrt(2).
# g30 = sympy.Matrix([[(2 - 3 * s) / 4, (2 + 3 * s) / 4]])
CURVE30 = bezier.Curve(np.array([
    [0.5, 0.5],
    [-0.25, 1.25],
]))
# g31 = 2 * g11 + sympy.Matrix([[1, 3]]) / 4
# g31 = sympy.Matrix([[(48 * s + 1) / 4, (67 - 64 * s) / 4]])
CURVE31 = bezier.Curve(np.array([
    [0.25, 16.75],
    [12.25, 0.75],
]))
# g32 = sympy.Matrix([[(7 * s - 2) / 8, -1 / 4]])
# NOTE: This is a degree-elevated line.
CURVE32 = bezier.Curve(np.array([
    [-0.25, -0.25],
    [0.1875, -0.25],
    [0.625, -0.25],
]))
# g33 = sympy.Matrix([[(1 - 2 * s) * (s - 1) / 8, -s]])
CURVE33 = bezier.Curve(np.array([
    [-0.125, 0.0],
    [0.0625, -0.5],
    [0.0, -1.0],
]))
# g34 = sympy.Matrix([[4, 3 * (2 * s + 5)]]) / 8
CURVE34 = bezier.Curve(np.array([
    [0.5, 1.875],
    [0.5, 2.625],
]))
# g35 = sympy.Matrix([[5 * (1 - 2 * s), 21]]) / 8
CURVE35 = bezier.Curve(np.array([
    [0.625, 2.625],
    [-0.625, 2.625],
]))
# g36 = sympy.Matrix([[(1 - s) / 2, -3 * (s + 1) / 8]])
CURVE36 = bezier.Curve(np.array([
    [0.5, -0.375],
    [0.0, -0.75],
]))
# g37 = sympy.Matrix([[(7 * s - 2) / 4, 7 * (s - 1) / 8]])
CURVE37 = bezier.Curve(np.array([
    [-0.5, -0.875],
    [1.25, 0.0],
]))


def make_plots(curve1, curve2, points, ignore_save=False, failed=True):
    if not CONFIG.running:
        return

    ax = curve1.plot(64)
    curve2.plot(64, ax=ax)
    ax.plot(points[:, 0], points[:, 1],
            marker='o', linestyle='None', color='black')
    ax.axis('scaled')
    _plot_helpers.add_plot_boundary(ax)

    if CONFIG.save_plot:
        if not ignore_save:
            CONFIG.save_fig()
    else:
        if failed:
            plt.title(CONFIG.current_test + ': failed')
        else:
            plt.title(CONFIG.current_test)
        plt.show()

    plt.close(ax.figure)


def curve_curve_check(curve1, curve2, s_vals, t_vals, points,
                      ignore_save=False):
    assert len(s_vals) == len(t_vals)
    assert len(s_vals) == len(points)

    intersections = _intersection_helpers.all_intersections(
        [(curve1, curve2)])
    assert len(intersections) == len(s_vals)

    info = six.moves.zip(intersections, s_vals, t_vals, points)
    for intersection, s_val, t_val, point in info:
        assert intersection.left is curve1
        assert intersection.right is curve2

        CONFIG.assert_close(intersection.s, s_val)
        CONFIG.assert_close(intersection.t, t_val)

        CONFIG.assert_close(intersection.point[0], point[0])
        CONFIG.assert_close(intersection.point[1], point[1])

        point_on1 = curve1.evaluate(s_val)
        CONFIG.assert_close(point_on1[0], point[0])
        CONFIG.assert_close(point_on1[1], point[1])

        point_on2 = curve2.evaluate(t_val)
        CONFIG.assert_close(point_on2[0], point[0])
        CONFIG.assert_close(point_on2[1], point[1])

    make_plots(curve1, curve2, points, ignore_save=ignore_save, failed=False)


def test_curves1_and_2():
    sq31 = np.sqrt(31.0)
    s_val0 = 0.0625 * (9.0 - sq31)
    s_val1 = 0.0625 * (9.0 + sq31)

    s_vals = np.array([s_val0, s_val1])
    t_vals = np.array([s_val1, s_val0])
    points = np.array([
        [s_val0, (16.0 + sq31) / 64.0],
        [s_val1, (16.0 - sq31) / 64.0],
    ])
    curve_curve_check(CURVE1, CURVE2, s_vals, t_vals, points)


def test_curves3_and_4():
    s_vals = np.array([0.25, 0.875])
    t_vals = np.array([0.75, 0.25])
    points = np.array([
        [0.75, 1.125],
        [2.625, 0.65625],
    ])
    curve_curve_check(CURVE3, CURVE4, s_vals, t_vals, points)


def test_curves1_and_5():
    s_vals = np.array([0.25, 0.75])
    t_vals = s_vals
    points = np.array([
        [0.25, 0.375],
        [0.75, 0.375],
    ])
    curve_curve_check(CURVE1, CURVE5, s_vals, t_vals, points)


def test_curves1_and_6():
    s_vals = np.array([0.5])
    t_vals = s_vals
    points = np.array([
        [0.5, 0.5],
    ])
    curve_curve_check(CURVE1, CURVE6, s_vals, t_vals, points)


def test_curves1_and_7():
    delta = 2.0 / np.sqrt(33.0)
    s_val0 = 0.5 - delta
    s_val1 = 0.5 + delta

    s_vals = np.array([s_val0, s_val1])
    t_vals = s_vals
    y_val = 17.0 / 66.0
    points = np.array([
        [s_val0, y_val],
        [s_val1, y_val],
    ])
    curve_curve_check(CURVE1, CURVE7, s_vals, t_vals, points)


def test_curves1_and_8():
    s_vals = np.array([0.25, 0.75])
    t_vals = s_vals
    points = np.array([
        [0.25, 0.375],
        [0.75, 0.375],
    ])
    curve_curve_check(CURVE1, CURVE8, s_vals, t_vals, points)


def test_curves1_and_9():
    s_vals = np.array([0.5])
    t_vals = np.array([2.0 / 3.0])
    points = np.array([
        [0.5, 0.5],
    ])
    curve_curve_check(CURVE1, CURVE9, s_vals, t_vals, points)


def test_curves10_and_11():
    s_vals = np.array([1.0 / 3.0])
    t_vals = np.array([0.5])
    points = np.array([
        [3.0, 4.0],
    ])
    curve_curve_check(CURVE10, CURVE11, s_vals, t_vals, points)


def test_curve12_self_crossing():
    left, right = CURVE12.subdivide()
    # Re-create left and right so they aren't sub-curves.
    left = bezier.Curve(left.nodes)
    right = bezier.Curve(right.nodes)

    delta = np.sqrt(5.0) / 3.0
    s_vals = np.array([1.0, 1.0 - delta])
    t_vals = np.array([0.0, delta])
    points = np.array([
        [-0.09375, 0.828125],
        [-0.25, 1.375],
    ])

    curve_curve_check(left, right, s_vals, t_vals, points)

    # Make sure the left curve doesn't cross itself.
    left1, right1 = left.subdivide()
    expected = right1.evaluate_multi(np.array([0.0]))
    curve_curve_check(bezier.Curve(left1.nodes),
                      bezier.Curve(right1.nodes),
                      np.array([1.0]),
                      np.array([0.0]),
                      expected, ignore_save=True)

    # Make sure the right curve doesn't cross itself.
    left2, right2 = right.subdivide()
    expected = right2.evaluate_multi(np.array([0.0]))
    curve_curve_check(bezier.Curve(left2.nodes),
                      bezier.Curve(right2.nodes),
                      np.array([1.0]),
                      np.array([0.0]),
                      expected, ignore_save=True)


def test_curves8_and_9():
    s_vals = np.array([0.5])
    t_vals = np.array([0.5])
    points = np.array([
        [0.5, 0.375],
    ])
    curve_curve_check(CURVE8, CURVE9, s_vals, t_vals, points)


def test_curves1_and_13():
    delta = 0.5 / np.sqrt(7.0)
    s_vals = np.array([0.5 - delta, 0.5 + delta, 0.0, 1.0])
    t_vals = s_vals
    points = np.array([
        [0.5 - delta, 3.0 / 7.0],
        [0.5 + delta, 3.0 / 7.0],
        [0.0, 0.0],
        [1.0, 0.0],
    ])
    curve_curve_check(CURVE1, CURVE13, s_vals, t_vals, points)


def test_curves14_and_15():
    s_vals = np.array([2.0 / 3.0])
    t_vals = np.array([1.0 / 3.0])
    points = np.array([
        [0.5, 0.5],
    ])
    with pytest.raises(NotImplementedError):
        curve_curve_check(CURVE14, CURVE15, s_vals, t_vals, points)

    make_plots(CURVE14, CURVE15, points)


def test_curves14_and_16():
    s_vals = np.array([3.0, 5.0]) / 6.0
    t_vals = np.array([1.0, 3.0]) / 6.0
    points = np.array([
        [0.375, 0.46875],
        [0.625, 0.46875],
    ])
    curve_curve_check(CURVE14, CURVE16, s_vals, t_vals, points)


def test_curves10_and_17():
    s_vals = np.array([1.0 / 3.0])
    t_vals = np.array([1.0])
    points = np.array([
        [3.0, 4.0],
    ])
    curve_curve_check(CURVE10, CURVE17, s_vals, t_vals, points)


def test_curves1_and_18():
    s_vals = np.array([1.0])
    t_vals = np.array([0.0])
    points = np.array([
        [1.0, 0.0],
    ])
    curve_curve_check(CURVE1, CURVE18, s_vals, t_vals, points)


def test_curves1_and_19():
    s_vals = np.array([1.0])
    t_vals = np.array([1.0])
    points = np.array([
        [1.0, 0.0],
    ])
    curve_curve_check(CURVE1, CURVE19, s_vals, t_vals, points)


def test_curves1_and_20():
    delta = np.sqrt(5.0) / 8.0
    s_vals = np.array([0.25, 0.375 - delta, 1.0, 0.375 + delta])
    t_vals = np.array([0.75, 0.625 - delta, 0.0, 0.625 + delta])
    points = np.array([
        [0.25, 0.375],
        [0.375 - delta, 0.3125 - 0.5 * delta],
        [1.0, 0.0],
        [0.375 + delta, 0.3125 + 0.5 * delta],
    ])
    curve_curve_check(CURVE1, CURVE20, s_vals, t_vals, points)


def test_curves20_and_21():
    sq5 = np.sqrt(5.0)
    s_vals = np.array([0.625 - 0.125 * sq5, 0.0, 0.75, 0.625 + 0.125 * sq5])
    t_vals = np.array([4.0 - sq5, 9.0, 3.0, 4.0 + sq5]) / 10.0
    points = np.array([
        [0.375 - 0.125 * sq5, 0.3125 - 0.0625 * sq5],
        [1.0, 0.0],
        [0.25, 0.375],
        [0.375 + 0.125 * sq5, 0.3125 + 0.0625 * sq5],
    ])
    curve_curve_check(CURVE20, CURVE21, s_vals, t_vals, points)


def test_curves21_and_22():
    sq5 = np.sqrt(5.0)
    s_vals = np.array([4.0 - sq5, 3.0, 9.0, 4.0 + sq5]) / 10.0
    t_vals = np.array([6.0 - sq5, 7.0, 1.0, 6.0 + sq5]) / 10.0
    points = np.array([
        [0.375 - 0.125 * sq5, 0.3125 - 0.0625 * sq5],
        [0.25, 0.375],
        [1.0, 0.0],
        [0.375 + 0.125 * sq5, 0.3125 + 0.0625 * sq5],
    ])

    # NOTE: We require a bit more wiggle room for these roots.
    with CONFIG.wiggle(12):
        curve_curve_check(CURVE21, CURVE22, s_vals, t_vals, points)


def test_curves10_and_23():
    s_vals = np.array([0.5])
    t_vals = np.array([3.0 / 10.0])
    points = np.array([
        [4.5, 4.5],
    ])
    curve_curve_check(CURVE10, CURVE23, s_vals, t_vals, points)


def test_curves1_and_24():
    # NOTE: This is two coincident curves, i.e. CURVE1 on
    #       [1/4, 1] is identical to CURVE24 on [0, 3/4].
    left_vals = CURVE1.evaluate_multi(
        np.linspace(0.25, 1.0, 2**8 + 1))
    right_vals = CURVE24.evaluate_multi(
        np.linspace(0.0, 0.75, 2**8 + 1))
    assert np.all(left_vals == right_vals)

    s_vals = np.array([0.25, 1.0])
    t_vals = np.array([0.0, 0.75])
    points = np.array([
        [0.25, 0.375],
        [1.0, 0.0],
    ])
    with pytest.raises(NotImplementedError):
        curve_curve_check(CURVE1, CURVE24, s_vals, t_vals, points)

    make_plots(CURVE1, CURVE24, points)


def test_curves15_and_25():
    # ctx = mpmath.MPContext()
    # ctx.prec = 200
    # ctx.polyroots([486, -3726, 13905, -18405, 6213, 1231])
    s_vals = np.array([float.fromhex('0x1.b7b348cf939b9p-1')])
    # ctx.polyroots([4, -16, 13, 25, -28, 4])
    t_vals = np.array([float.fromhex('0x1.bf3536665a0cdp-1')])

    # Slightly more accurate than (3 s + 1) / 4.
    x_val = float.fromhex('0x1.c9c6769baeb4ap-1')
    # identical to (9 s^2 - 6 s + 5) / 8 after rounding.
    y_val = float.fromhex('0x1.9f09401c281ddp-1')
    points = np.array([
        [x_val, y_val],
    ])
    curve_curve_check(CURVE15, CURVE25, s_vals, t_vals, points)


def test_curves11_and_26():
    sq7 = np.sqrt(7.0)
    s_vals = np.array(
        [24.0, 24.0 - 7.0 * sq7, 24.0 + 7.0 * sq7]) / 48.0
    t_vals = np.array([3.0, 3.0 - sq7, 3.0 + sq7]) / 6.0
    points = np.array([
        [72.0, 96.0],
        [72.0 - 21.0 * sq7, 96.0 + 28.0 * sq7],
        [72.0 + 21.0 * sq7, 96.0 - 28.0 * sq7],
    ]) / 24.0

    # NOTE: We require a bit more wiggle room for these roots.
    with CONFIG.wiggle(25):
        curve_curve_check(CURVE11, CURVE26, s_vals, t_vals, points)


def test_curves8_and_27():
    s_val1, s_val2, _ = runtime_utils.real_roots(
        [17920, -29760, 13512, -1691])
    s_vals = np.array([s_val2, s_val1])
    t_val1, t_val2, _ = runtime_utils.real_roots([35, -60, 24, -2])
    t_vals = np.array([t_val2, t_val1])
    points = np.array([
        [s_val2, 0.375],
        [s_val1, 0.375],
    ])

    # NOTE: We require a bit more wiggle room for these roots.
    with CONFIG.wiggle(42):
        curve_curve_check(CURVE8, CURVE27, s_vals, t_vals, points)


def test_curves28_and_29():
    # NOTE: Though curves 1 and 6 successfully intersect (at a point
    #       of tangency), the rotated equivalents do not.
    s_vals = np.array([0.5])
    t_vals = s_vals
    points = np.array([
        [0.0, 1.0],
    ])
    with pytest.raises(NotImplementedError):
        curve_curve_check(CURVE28, CURVE29, s_vals, t_vals, points)

    make_plots(CURVE28, CURVE29, points)


def test_curves29_and_30():
    s_vals = np.array([0.5])
    t_vals = np.array([2.0 / 3.0])
    points = np.array([
        [0.0, 1.0],
    ])
    curve_curve_check(CURVE29, CURVE30, s_vals, t_vals, points)


def test_curves8_and_23():
    s_vals = np.zeros((0,))
    t_vals = s_vals
    points = np.zeros((0, 2))
    curve_curve_check(CURVE8, CURVE23, s_vals, t_vals, points)


def test_curves11_and_31():
    s_vals = np.zeros((0,))
    t_vals = s_vals
    points = np.zeros((0, 2))
    curve_curve_check(CURVE11, CURVE31, s_vals, t_vals, points)


def test_curves32_and_33():
    s_vals = np.array([13.0 / 56.0])
    t_vals = np.array([0.25])
    points = np.array([
        [-0.046875, -0.25],
    ])

    # NOTE: We allow less wiggle room for this intersection.
    with CONFIG.wiggle(4):
        curve_curve_check(CURVE32, CURVE33, s_vals, t_vals, points)


def test_curves34_and_35():
    s_vals = np.array([1.0])
    t_vals = np.array([1.0 / 10.0])
    points = np.array([
        [0.5, 2.625],
    ])

    curve_curve_check(CURVE34, CURVE35, s_vals, t_vals, points)


def test_curves36_and_37():
    s_vals = np.array([0.0])
    t_vals = np.array([4.0 / 7.0])
    points = np.array([
        [0.5, -0.375],
    ])

    curve_curve_check(CURVE36, CURVE37, s_vals, t_vals, points)


if __name__ == '__main__':
    CONFIG.run(globals())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper for B |eacute| zier Curves, Triangles, and Higher Order Objects.

Intended to perform basic operations on B |eacute| zier objects such
as intersections, length/area/etc. computations, subdivision,
implicitization and other relevant information.

Plotting utilities will also be provided.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:
"""

from bezier.curve import Curve
from bezier.curved_polygon import CurvedPolygon
from bezier.surface import Surface


__all__ = [
    'Curve',
    'CurvedPolygon',
    'Surface',
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

"""Utilities for running functional tests as scripts."""


import argparse
import contextlib
import inspect
import os
import types

import matplotlib.pyplot as plt
import numpy as np
try:
    import seaborn  # pylint: disable=unused-import
except ImportError:
    pass
import six

from bezier import _helpers


_FNL_TESTS_DIR = os.path.dirname(__file__)
_DOCS_DIR = os.path.abspath(
    os.path.join(_FNL_TESTS_DIR, '..', 'docs'))
IMAGES_DIR = os.path.join(_DOCS_DIR, 'images')


def _start_line(func):
    """Get the start line (in source) of a function.

    Args:
        func (~types.FunctionType): A Python function object.

    Returns:
        int: The start line (in source).
    """
    _, line = inspect.getsourcelines(func)
    return line


def get_parser():
    """Create a command line argument parser.

    Returns:
        argparse.ArgumentParser: An argument parser for functional tests.
    """
    parser = argparse.ArgumentParser(
        description='Run functional tests.')
    parser.add_argument('--save-plot', dest='save_plot',
                        action='store_true')
    return parser


def real_roots(coeffs):
    """Get real roots of a polynomial.

    Args:
        coeffs (List[Float]): List of polynomial coefficients.

    Returns:
        numpy.ndarray: The (sorted) real roots of the polynomial.
    """
    all_roots = np.roots(coeffs)
    filtered = all_roots[all_roots.imag == 0.0].real
    return np.sort(filtered)


class Config(object):  # pylint: disable=too-few-public-methods
    """Run-time configuration.

    This is a mutable stand-in to allow test set-up to modify
    global state.
    """

    def __init__(self):
        self.running = False
        self.save_plot = False
        self.current_test = None
        self.parser = get_parser()
        self._wiggle = 8

    def _run(self, mod_globals):
        """Run all tests, in source order.

        Args:
            mod_globals (dict): The globals from a module.
        """
        found = []
        for key, value in six.iteritems(mod_globals):
            if not key.lower().startswith('test'):
                continue
            if not isinstance(value, types.FunctionType):
                continue
            found.append(value)

        found.sort(key=_start_line)

        for func in found:
            self.current_test = func.__name__
            func()

    @contextlib.contextmanager
    def wiggle(self, wiggle):
        """Make the context use a temporary wiggle room.

        Args:
            wiggle (int): The temporary amount of wiggle room.
        """
        old_wiggle = self._wiggle
        try:
            self._wiggle = wiggle
            yield self
        finally:
            self._wiggle = old_wiggle

    def assert_close(self, approximated, exact):
        """Assert two values are close, with the local configuration in place.

        Args:
            approximated (float): The value that was computed.
            exact (float): The expected value.
        """
        assert _helpers.n_bits_away(
            exact, approximated, num_bits=self._wiggle)

    def run(self, mod_globals):
        """Run all tests, in source order.

        Args:
            mod_globals (dict): The globals from a module.
        """
        running = self.running
        save_plot = self.save_plot
        try:
            self.running = True
            args = self.parser.parse_args()
            self.save_plot = args.save_plot
            self._run(mod_globals)
        finally:
            self.running = running
            self.save_plot = save_plot
            self.current_test = None

    def save_fig(self):
        """Save the current figure.

        Uses the ``current_test`` for the filename and puts it
        in the ``${GIT_ROOT}/docs/images`` directory.
        """
        filename = '{}.png'.format(self.current_test)
        path = os.path.join(IMAGES_DIR, filename)
        plt.savefig(path, bbox_inches='tight')
        print('Saved {}'.format(filename))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common features of B |eacute| zier shapes.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:
"""

import numpy as np


class Base(object):
    """Base shape object.

    Args:
        nodes (numpy.ndarray): The control points for the shape.
            Must be a 2D array, where the rows are the nodes and the
            columns correspond to each dimension the shape occurs in.
        _copy (bool): Flag indicating if the nodes should be copied before
            being stored. Defaults to :data:`True` since callers may
            freely mutate ``nodes`` after passing in.

    Raises:
        ValueError: If the ``nodes`` are not 2D.
        ValueError: If the ``degree`` is less than ``1``.
    """

    def __init__(self, nodes, _copy=True):
        if nodes.ndim != 2:
            raise ValueError('Nodes must be 2-dimensional, not',
                             nodes.ndim)

        num_nodes, dimension = nodes.shape
        degree = self._get_degree(num_nodes)
        if degree < 1:
            raise ValueError('Shape must be at least degree 1')

        self._degree = degree
        self._dimension = dimension
        if _copy:
            self._nodes = nodes.copy()
        else:
            self._nodes = nodes

    @staticmethod
    def _get_degree(num_nodes):
        """Get the degree of the current shape.

        .. note::

           Subclasses are expected to implement this method.

        Args:
            num_nodes (int): The number of nodes provided.

        Returns:
            int: The degree of the current shape.
        """
        return num_nodes

    @property
    def degree(self):
        """int: The degree of the current shape."""
        return self._degree

    @property
    def dimension(self):
        r"""int: The dimension that the shape lives in.

        For example, if the shape lives in :math:`\mathbf{R}^3`, then
        the dimension is ``3``.
        """
        return self._dimension

    @property
    def nodes(self):
        """numpy.ndarray: The nodes that define the current shape."""
        return self._nodes.copy()

    # NOTE: We disable the docstring return type check because
    #       subclasses won't be able to refer to this private base.
    def copy(self):  # pylint: disable=missing-returns-doc
        """Make a copy of the current shape.

        Returns:
            Instance of the current shape.
        """
        return self.__class__(self._nodes, _copy=True)

    def __eq__(self, other):
        """Check equality against another shape.

        Returns:
            bool: Boolean indicating if the shapes are the same.
        """
        # First check if both are instances of an identical class.
        if getattr(other, '__class__', None) != self.__class__:
            return False
        if self.degree != other.degree:
            return False
        if self.dimension != other.dimension:
            return False
        other_nodes = other._nodes  # pylint: disable=protected-access
        return np.all(self._nodes == other_nodes)

    def __ne__(self, other):
        """Check inequality against another shape.

        Returns:
            bool: Boolean indicating if the shapes are not the same.
        """
        return not self.__eq__(other)

    def __repr__(self):
        """Representation of current object.

        Returns:
            str: Object representation.
        """
        return '<{} (degree={:d}, dimension={:d})>'.format(
            self.__class__.__name__, self.degree, self.dimension)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup file for bezier."""

import os

from setuptools import find_packages
from setuptools import setup


VERSION = '0.2.1'
PACKAGE_ROOT = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(PACKAGE_ROOT, 'README.rst.template')) as file_obj:
    TEMPLATE = file_obj.read()

README = TEMPLATE.format(
    pypi='',
    pypi_img='',
    versions='',
    versions_img='',
    rtd_version=VERSION,
    coveralls_branch=VERSION,
)

REQUIREMENTS = (
    'enum34',
    'matplotlib >= 1.5.3',
    'numpy >= 1.11.2',
    'six >= 1.9.0',
)
DESCRIPTION = (
    u'Helper for B\u00e9zier Curves, Triangles, and Higher Order Objects')


setup(
    name='bezier',
    version=VERSION,
    description=DESCRIPTION,
    author='Danny Hermes',
    author_email='daniel.j.hermes@gmail.com',
    long_description=README,
    scripts=(),
    url='https://github.com/dhermes/bezier',
    packages=find_packages(),
    license='Apache 2.0',
    platforms='Posix; MacOS X; Windows',
    include_package_data=True,
    zip_safe=True,
    install_requires=REQUIREMENTS,
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet',
    ),
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic geometry and floating point helpers."""


import numpy as np


_EPS = 2.0**(-40)


def vector_close(vec1, vec2, eps=_EPS):
    r"""Checks that two vectors are equal to some threshold.

    Does so by computing :math:`s_1 = \|v_1\|_2` and
    :math:`s_2 = \|v_2\|_2` and then checking if

    .. math::

       \|v_1 - v_2\|_2 \leq \varepsilon \min(s_1, s_2)

    where :math:`\varepsilon = 2^{-40} \approx 10^{-12}` is a fixed
    threshold. In the rare case that one of ``vec1`` or ``vec2`` is
    the zero vector (i.e. when :math:`\min(s_1, s_2) = 0`) instead
    checks that the other vector is close enough to zero:

    .. math::

       \|v_1\|_2 = 0 \Longrightarrow \|v_2\|_2 \leq \varepsilon

    .. note::

       This function assumes that both vectors have finite values,
       i.e. that no NaN or infinite numbers occur. NumPy provides
       :func:`np.allclose` for coverage of **all** cases.

    Args:
        vec1 (numpy.ndarray): First vector for comparison.
        vec2 (numpy.ndarray): Second vector for comparison.
        eps (float): Error threshold. Defaults to :math:`2^{-40}`.

    Returns:
        bool: Flag indicating if they are close to precision.
    """
    size1 = np.linalg.norm(vec1, ord=2)
    size2 = np.linalg.norm(vec2, ord=2)
    if size1 == 0:
        return size2 <= eps
    elif size2 == 0:
        return size1 <= eps
    else:
        upper_bound = eps * min(size1, size2)
        return np.linalg.norm(vec1 - vec2, ord=2) <= upper_bound


def in_interval(value, start, end):
    """Checks if a ``value`` is an interval (inclusive).

    .. note::

       The current implementation does the most basic check,
       however, in the future, a more generic check may be desired
       that allows wiggle room around the endpoints to account
       for round-off.

    Args:
        value (float): The value to check.
        start (float): The (inclusive) start of the interval.
        end (float): The (inclusive) end of the interval.

    Returns:
        bool: Indicating if the value is in the interval.
    """
    return start <= value <= end


def bbox(nodes):
    """Get the bounding box for set of points.

    Args:
       nodes (numpy.ndarray): A set of points.

    Returns:
        Tuple[float, float, float, float]: The left, right,
        bottom and top bounds for the box.
    """
    left, bottom = np.min(nodes, axis=0)
    right, top = np.max(nodes, axis=0)
    return left, right, bottom, top


def contains(nodes, x_val, y_val):
    r"""Predicate indicating if a point is within a bounding box.

    Args:
       nodes (numpy.ndarray): A set of points.
       x_val (float): The :math:`x`-coordinate of the point.
       y_val (float): The :math:`y`-coordinate of the point.

    Returns:
        bool: Indicating containment.
    """
    left, right, bottom, top = bbox(nodes)
    if not in_interval(x_val, left, right):
        return False
    if not in_interval(y_val, bottom, top):
        return False
    return True


def contains_nd(nodes, point):
    r"""Predicate indicating if a point is within a bounding box.

    Like :func:`contains` but supports points in arbitrary dimension.
    Unlike :func:`contains`, this function directly uses ``<=`` and
    ``>=`` for comparison (:func:`contains` uses :func:`in_interval`).

    Args:
       nodes (numpy.ndarray): A set of points.
       point (numpy.ndarray): A 1D NumPy array representing a point
           in the same dimension as ``nodes``.

    Returns:
        bool: Indicating containment.
    """
    min_vals = np.min(nodes, axis=0)
    if not np.all(min_vals <= point):
        return False

    max_vals = np.max(nodes, axis=0)
    if not np.all(point <= max_vals):
        return False

    return True


def cross_product(vec0, vec1):
    r"""Compute the cross-product of vectors in :math:`\mathbf{R}^2`.

    Utilizes the fact that

    .. math::

       \left[\begin{array}{c} A \\ B \\ 0 \end{array}\right] \times
           \left[\begin{array}{c} C \\ D \\ 0 \end{array}\right] =
           \left[\begin{array}{c} 0 \\ 0 \\ AD - BC \end{array}\right]

    and just returns the :math:`z` component.

    Args:
        vec0 (numpy.ndarray): A vector as a 1x2 NumPy array.
        vec1 (numpy.ndarray): A vector as a 1x2 NumPy array.

    Returns:
        float: The cross-product (or rather, its :math:`z` component).
    """
    return vec0[0, 0] * vec1[0, 1] - vec0[0, 1] * vec1[0, 0]


def n_bits_away(value1, value2, num_bits=1):
    r"""Determines if ``value1`` is within ``n`` bits of ``value2``.

    Uses ``np.spacing`` to determine the unit of least precision (ULP)
    for ``value1`` and then checks that the different between the values
    does not exceed ``n`` ULPs.

    When ``value1=0`` or ``value2=0``, we instead check that the other is
    less than :math:`2^{-40}` in magnitude.

    Args:
        value1 (float): The first value that being compared.
        value2 (float): The second value that being compared.
        num_bits (Optional[int]): Defaults to ``1``.

    Returns:
        bool: Predicate indicating if the values agree to ``n`` bits.
    """
    if value1 == 0.0:
        return abs(value2) < _EPS
    elif value2 == 0.0:
        return abs(value1) < _EPS
    else:
        local_epsilon = np.spacing(value1)  # pylint: disable=no-member
        return abs(value1 - value2) <= num_bits * abs(local_epsilon)
